#!/usr/bin/env perl
# weave-pbo-c.pl - weave pbo-c source file

# call as: weave-pbo-c <source_file>
# input: <source_file>
# output: <source_file>.c <source_file>.opt <source_file>.log

# (C) Holger H. Hoos, University of British Columbia 2010
# version 0.6.06 (2012/09/10) - Quinn Hsu
# - minor version update to keep in sync with other weavers

use FindBin;
use lib ($FindBin::Bin);
use strict;

#use warnings;

use PBO::WeaverModule qw(:ALL);
use File::Basename;
use File::Path qw(make_path);

my $weaverId = "weave-pbo-c 0.6.06";

$| = 1;

$ext = "\.pbo-j";

# deal with instantiations:


my %typPC = (
	"bool"		=> "boolean",
	"boolean"	=> "boolean",
	"int"		=> "int",
	"integer"	=> "int",
	"float"		=> "double",
	"real"		=> "double",
	"char"		=> "char",
	"character"	=> "char",
	"str"		=> "String",
	"string"	=> "String"
);

# TODO: support other param types

#my %paramDeclLine   = ();    # used for collecting info on param decl
#my %paramAccLines   = ();    # used for collecting info on param access
#my %choiceLines     = ();    # used for collecting info on choices
#my %choiceInstNames = ();    # used for collecting info on choice instance names
#my %logLines		= ();	 # used for collecting info on logs

sub generateSupportFile {
	# modified in 0.5
	
	# generate PBO_Support.java (assumes that parameter and choice declarations have been collected)
	my $sourcename = join ", ", @inputFiles;
	
	my $directory = "ca/ubc/beta/pbo";
	
	unless (-d $directory) {
	    make_path $directory
	}

	print "\n\nGenerating $directory/PBOSupport.java from $sourcename ...\n";
	print $LOG "\nGenerating $directory/PBOSupport.java from $sourcename ...\n";

	open( DEST, ">","$directory/PBOSupport.java" ) or die("WARNING: Could not open file $directory/PBOSupport.java: $!\n");
	
	print DEST <<"END";
package ca.ubc.beta.pbo;

import java.util.HashMap;
import java.io.BufferedReader;
import java.io.FileReader;
import java.util.regex.*;
import java.io.IOException;

\@SuppressWarnings("all")
public class PBOSupport {
	
END
	print DEST "	//parameters\n";
	
	for my $i (0 .. $#params) {
		my $tc = $typPC{$params[$i]{type}};
		if ($tc ne "") {
			print DEST "	public final $tc $params[$i]{name};\n";
			
			if (exists $params[$i]{lbound} && exists $params[$i]{rbound}) {
				print DEST "	private final $tc $params[$i]{name}"."_lbound;\n";
				print DEST "	private final $tc $params[$i]{name}"."_rbound;\n";
			}
			
			if (exists $params[$i]{categories}) {
				print DEST "	private final $tc\[\] $params[$i]{name}"."_categories;\n";
			}
		} else {
			print DEST "\nERROR: Unknown type'$params[$i]{type}'\n";
		}
	}
	
	print DEST "	//choices\n";
	
	my %seenChoices = ();
	my @uniqueChoiceNames = map {$_->{name}} (grep { !$seenChoices{$_->{name}}++ } @choices);
	
	for my $choiceName (@uniqueChoiceNames) {
		print DEST "	public final int $choiceName;\n";
	}
	
	print DEST "	// choices with ids\n";
	
	my %seenChoicesWithIDs = ();
	my @choicesWithIDs = grep {!$seenChoicesWithIDs{$_}++} (map { "$_->{name}\_$_->{value}" } (grep {$_->{value} ne ""} @choices));
	for my $choice_id_pair (@choicesWithIDs) {
		print DEST "	public final int $choice_id_pair;\n";
	}
	
	print DEST <<"END";
	
	private static HashMap<String,String> pboArguments;
	private static final Pattern nameValue = Pattern.compile(\"--(.+)=(.+)\");
	private static String[] commandLineArguments = null;
	private static String callMainMethodClass = null;
	
	private static PBOSupport pbo_support = null;
	
	public long startTime;
	
	private static void setCommandLineArguments(String[] args) {
		commandLineArguments = args;
		
		// do a pass through the command line arguments
		// so that we can call the project's main method
		// without having to create an instance of this
		// singleton first
		
		pbo_support = new PBOSupport();
	}
	
	public static PBOSupport getInstance() {
		if (pbo_support == null) {
			throw new IllegalStateException("The PBOSupport class has not been intialized.  The PBOSupport main method should be the point of entry for project.");
		}
		
		return pbo_support;
	}
	
	private PBOSupport() {
		pboArguments = new HashMap<String,String>();
		
		// we cannot initialize these variables within a method
		// so we must do them in the constructor
		
END
	print DEST "		// params with defaults\n\n";
	
	my @paramsWithDefaults = grep {$_->{default} == 1} @params;
	for my $i (0 .. $#paramsWithDefaults) {
		my $name = $paramsWithDefaults[$i]{name};
		
		if (!(exists $instList{$name})) {
		    my $value = $paramsWithDefaults[$i]{value};
		    ($value) = $value =~ /(?|(?=")"(.+)"|(?=')'(.+)'|(.+))/g;
		
		    print DEST "		pboArguments.put(\"$name\",\"$value\");\n";
		}
	}
	
	print DEST "\n		// params with ranges\n\n";
	
	my @paramsWithRanges = grep {exists $_->{lbound} && $_->{rbound}} @params;
	for my $param (@paramsWithRanges) {
	    if(!(exists $instList{$param->{name}})) {
		    print DEST "		$param->{name}\_lbound = $param->{lbound};\n";
		    print DEST "		$param->{name}\_rbound = $param->{rbound};\n";
		}
	}
	
	print DEST "\n		// params with categories\n\n";
	
	my @paramsWithCategories = grep {exists $_->{categories}} @params;
	for my $param (@paramsWithCategories) {
	    if (!(exists $instList{$param->{name}})) {
		    print DEST "		$param->{name}\_categories = new ".$typPC{$param->{type}}."[] {".join(",", @{$param->{categories}})."};\n";
		    # new ".$typPC{$param->{type}}."[".@{$param->{categories}}."]
		}
		
	}
	
	# this needs to be generated here, but that doesn't mean i have to 
	# like it
	
	#@invInstNames{@choicesWithIDs} = (1 .. $#choicesWithIDs+1);
	#
	#my @choicesWithIDsAndValues = map { "$choicesWithIDs[$_] = ".($_+1) } (0 .. $#choicesWithIDs);
	#for my $choice (@choicesWithIDsAndValues) {
	#	print DEST "		$choice;\n";
	#}
	
	print DEST "\n		// choices with ids\n\n";
	
	foreach my $choice (keys %invInstNames) {
	    if (!(exists $instList{$choice})) {
		    foreach my $id (keys %{$invInstNames{$choice}}) {
			    print DEST "		$choice\_$id=$invInstNames{$choice}{$id};\n";
		    }
		}
	}
	
	print DEST <<"END";
		
		
		// parse the command line and optional options file BEFORE
		// we validate that all the pbo constructs have been initialized
		
		try { parseCommandLine(commandLineArguments); }
		catch (Exception e) {
			System.out.println("PBOSupport: something went wrong with reading the options file");
			e.printStackTrace();
			System.exit(-1);
		}
		
		pbo_check_param_values();
		
		// we will be able to tell, at weaving time, what these
		// types of variables are, so there is no concern here to
		// record these types in some data structure.
		//
		// furthermore, we test out parse* BEFORE we try and 
		// assign it to the variable so that we can warn the 
		// user of malformed input

END
	print DEST "		// parameters\n";
	for my $i (0 .. $#params) {
		my $tc = $typPC{$params[$i]{type}};
		if ($tc eq "int") {
			print DEST <<"END";
		try {  Integer.parseInt(pboArguments.get(\"$params[$i]{name}\")); } catch (NumberFormatException e) { System.out.println(\"Supplied value for $params[$i]{name} must be a long integer\"); System.exit(-1); }
		$params[$i]{name} = Integer.parseInt(pboArguments.get(\"$params[$i]{name}\"));
END
		}
		if ($tc eq "double") {
			print DEST <<"END";
		try { Double.parseDouble(pboArguments.get(\"$params[$i]{name}\")); } catch (NumberFormatException e) { System.out.println(\"Supplied value for $params[$i]{name} must be a double or float\"); System.exit(-1); }
		$params[$i]{name} = Double.parseDouble(pboArguments.get(\"$params[$i]{name}\"));
END
		}
		if ($tc eq "boolean") {
			print DEST "		$params[$i]{name} = parseBoolean(pboArguments.get(\"$params[$i]{name}\"));\n";
		}
		if ($tc eq "char") {
			print DEST "		$params[$i]{name} = pboArguments.get(\"$params[$i]{name}\").charAt(0);\n";
		}
		if ($tc eq "String") {
			print DEST "		$params[$i]{name} = pboArguments.get(\"$params[$i]{name}\");\n";
		}
	}
	
	for my $choice (@uniqueChoiceNames) {
		print DEST <<"END";
		try { Integer.parseInt(pboArguments.get("$choice")); } catch (NumberFormatException e) { System.out.println("ERROR: "+pboArguments.get("$choice")+" is not a valid option for choice --$choice."); System.exit(-1); }
		$choice = Integer.parseInt(pboArguments.get(\"$choice\"));
END
	}
		
	print DEST <<"END";
		callMainMethodClass = pboArguments.get("javaClass");
		
		pbo_check_param_ranges();
		pbo_check_param_categories();
		pbo_check_choice_categories();
		
	}
	
	private void pbo_check_param_values() {
		// we assume that parseCommandLine has been called already
		
		// we can know at weaving time which params/choices do
		// not have defaults, so we force the user to provide
		// values for them first
END
	
	my @paramsWithNoDefaults = map { "\"$_->{name}\"" } (grep {$_->{default} == 0} @params);
	my $paramsWithNoDefaultsString = join ",", @paramsWithNoDefaults;
	
	my $choicesString = join ",", (map {"\"$_\""} @uniqueChoiceNames);
	
	print DEST "		String[] defaultlessParams  = {$paramsWithNoDefaultsString};\n";
	print DEST "		String[] choices			= {$choicesString};\n";
	print DEST "		String mainMethodClassString= \"javaClass\";\n";

	print DEST <<"END";
		for(int i=0;i<defaultlessParams.length; i++) {
			String param = defaultlessParams[i];
			
			if (!pboArguments.containsKey(param)) {
				System.out.format("ERROR: Parameter --%s must be set.%n",param);
				System.exit(-1);
			}
		}
		
		for(int i=0; i<choices.length; i++) {
			String choice = choices[i];
			
			if(!pboArguments.containsKey(choice)) {
				System.out.format("ERROR: Choice --%s must be set%n",choice);
				System.exit(-1);
			}
		}
		
		if (!pboArguments.containsKey(mainMethodClassString)) {
			System.out.println("ERROR: Project entry point --javaClass must be set.");
			System.exit(-1);
		}
	}
	
	private void pbo_check_param_ranges() {
END
	
	for my $param (@paramsWithRanges) {
		print DEST <<"END";
		if(!($param->{name}\_lbound <= $param->{name} || $param->{name} <= $param->{name}\_rbound)) {
			System.out.println("ERROR: Parameter --$param->{name} must be within bounds [$param->{lbound},$param->{rbound}]");
			System.exit(-1);
		}
END
	}

	print DEST <<"END";
	}
	
	private void pbo_check_param_categories() {
END
	
	for my $param (@paramsWithCategories) {
		if (grep { $param->{type} eq $_ } ("str", "string")) {
			my $categories = join ",", @{$param->{categories}};
			$categories =~ s/"/\\"/g;
			
			my $condition = join "||", (map { "$param->{name}.equals($_)" } @{$param->{categories}});
			print DEST <<"END";
		if(!($condition)) {
			System.out.println("ERROR: Parameter --$param->{name} must be one of \{$categories\}");
			System.exit(-1);
		}
END
		} else {
			my $categories = join ",", @{$param->{categories}};
			my $condition = join "&&", (map {"($param->{name}\_categories\[$_\] != $param->{name})"} (0 .. (@{$param->{categories}}-1)) );
			print DEST <<"END"
		if($condition) {
			System.out.println("ERROR: Parameter --$param->{name} must be one of \{$categories\}");
			System.exit(-1);
		}
END
		}
	}

	print DEST <<"END";
	}
	
	private void pbo_check_choice_categories() {
END
	my %seenUniqueChoiceNamesWithIDs = ();
	my @uniqueChoiceNamesWithIDs = map {$_->{name}} (grep {!$seenUniqueChoiceNamesWithIDs{$_->{name}}++} (grep {$_->{value} ne ""} @choices));
	
	# the problem here is that there does not exist a mapping from 
	# valid choice ids to their int counter parts
	
	my %uniqueChoiceNamesToNameInts = map { 
		my $name = $_; 
		$name => [values %{$invInstNames{$name}}]
	} @uniqueChoiceNamesWithIDs;
	
	for my $choice (@uniqueChoiceNamesWithIDs) {
		my @ints = sort @{$uniqueChoiceNamesToNameInts{$choice}};
		my @names = keys (%{ $invInstNames{$choice} });
		my $intString = join ",", (@ints,@names);
		my $condition = join "||", (map {"($choice == $_)"} @ints);
		print DEST <<"END";
		if (!($condition)) {
			System.out.println("ERROR: Choice --$choice must be one of {$intString}");
			System.exit(-1);
		}
END
	}

	print DEST <<"END";
	}
	
	private static void parseCommandLine(String[] args) throws IOException {
		// holds all the command line instantiations
		HashMap<String,String> commandLineArguments = new HashMap<String,String>();
		HashMap<String,String> optionsFileArguments = new HashMap<String,String>();
END

	my $num_choices = scalar (keys %invInstNames);
	print DEST	"		HashMap<String,HashMap<String,Integer>> choiceNamedInstances = new HashMap<String,HashMap<String,Integer>>($num_choices);\n";
	
	# modified in 0.6.04
	# automatic translation of choice named instances to their
	# integer counterparts.
	
	foreach my $choice (keys %invInstNames) {
		my $num_choice_names = scalar (keys %{ $invInstNames{$choice} });
			print DEST "		choiceNamedInstances.put(new String(\"$choice\"),new HashMap<String,Integer>($num_choice_names) {{ ".join(" ", map { "put(new String(\"$_\"),new Integer($invInstNames{$choice}{$_}));" } (keys %{ $invInstNames{$choice} })) ." }});\n";
	
	}

	print DEST <<"END";
		
		for(String arg: args) {
			if (arg.equals("--")) break;
			
			// we do not expect to see more than one of these
			if (arg.startsWith("@")) {
				String optFile = arg.substring(1); // chop @
				optionsFileArguments = parseOptionsFile(optFile);
			} else if (arg.startsWith("--")) {
				Matcher matcher = nameValue.matcher(arg);
				if(matcher.find()) {
					String name = matcher.group(1);
					String value = matcher.group(2);
					
					if (choiceNamedInstances.containsKey(name) && ((HashMap<String,Integer>)choiceNamedInstances.get(name)).containsKey(value)) {
						Integer val = ((HashMap<String,Integer>)choiceNamedInstances.get(name)).get(value);
						commandLineArguments.put(name,val.toString());
					} else {
						commandLineArguments.put(name,value);
					}
				}
			} else {
				System.out.format("WARNING: Unrecognized argument format: %s. Skipping.%n", arg);
			}
			
		}
		
		// because putAll overrides old values with new values for the
		// same key we simply abuse that fact.
		//
		// because we prioritize command line argument values over option
		// file values, we put the command line arguments last
		
		pboArguments.putAll(optionsFileArguments);
		pboArguments.putAll(commandLineArguments);
	}
	
	private static HashMap<String,String> parseOptionsFile(String pathToFile) throws IOException {
		HashMap<String,String> optionsFileArguments = new HashMap<String,String>();
		BufferedReader optionsFile = null;
		
		// do not greedily grab until we see '=' because choices with
		// ids have lines with two '='s in them.
		
		Pattern nameValue = Pattern.compile("\\\\s*--(.+?)\\\\s*=\\\\s*(.+)");
		
		try {
			optionsFile = new BufferedReader(new FileReader(pathToFile));
			String line = null;
			
			while((line = optionsFile.readLine()) != null) {
				// no use doing an expensive regex if we don't have to
				if (!line.startsWith(";")) {
					Matcher matcher = nameValue.matcher(line);
					if(matcher.find()) {
						String name = matcher.group(1);
						String value = matcher.group(2);
						
						// this is a really bad hack to get around the fact that
						// lines about choices with ids have a ; comment at the end
						// 
						// and java's regex doesn't support conditionals within the
						// regular expression.
						//
						// ex.
						// --processing=4 ; = alternative
						//
						// ideal the regular expression is
						// \s*--(.+?)=(.+)\s*(?(?=;)(?:.+))
						//
						// which means optional start of white space
						// --(name)=(value)
						// followed by optional white space
						// if followed by ; then ignore till end of line
						// 
						// the draw back to this simple regular expression is that if
						// value turns out to be a string or character that contains
						// ; in it, this regular expression will truncate everything after
						// the semicolon within the string.  i'm hoping the chances 
						//of that are small, and we can fix it if it ever arises.
						
						if (value.lastIndexOf(";") != -1) {
							value = value.substring(0, value.lastIndexOf(";")).trim();
						}
						
						// when parsing the opt file, characters like " and ' stick
						// around even though they do NOT when parsing the command line
						//
						// therefore, we need to remove these beginning and end characters
						
						if (value.startsWith(\"\\"\") && value.endsWith(\"\\"\")) {
							value = value.substring(1, value.length()-1);
						}
						
						if (value.startsWith(\"'\") && value.endsWith(\"'\")) {
							value = value.substring(1,value.length()-1);
						}
						
						optionsFileArguments.put(name,value);
					}
				}			
			}
		} finally {
			optionsFile.close();
		}
		
		return optionsFileArguments;
	}
	
	private static String[] buildMainProgramArguments(String[] args) {
		String[] mainArguments = null;
		boolean shouldCopy = false;
		int copyIndex=0;
		
		// unlike C, args does not contain the name of the program
		// as its first argument
		for(int i=0; i<args.length; i++) {
			// arguments to the main program are only supposed to 
			// come after "--"
			if (args[i].equals("--")) {
				shouldCopy=true;
				mainArguments = new String[args.length-i-1];
				
				// we do NOT want to fall through to the if
				// statement below since we do NOT want to
				// copy over "--" as well
				
				continue;
			}
			
			if (shouldCopy) { mainArguments[copyIndex++] = args[i]; }
		}	
		
		// if we don't find "--", then no main program arguments were
		// supplied, but Java specs requires us to at least give it
		// an empty array
		
		if (mainArguments == null) { return new String[0]; }
		
		return mainArguments;
	}
	
	// choices should be represented by booleans
	// here, we default to false unless one of "true", "1", "active" or "on"
	// is the keyword supplied
	// therefore, we do not throw a CouldNotParse error.
	public static boolean parseBoolean(String bool) {
		if (bool.equals("true")||bool.equals("1")||bool.equals("active")||bool.equals("on")) {
			return true;
		}
		
		return false;
	}
	
	// main point of entry for a PbO enriched source
	public static void main(String[] args) throws Throwable {
		setCommandLineArguments(args);
		
		String[] passThroughArguments = buildMainProgramArguments(args);
		
		getInstance().startTime = System.currentTimeMillis();
		
		try {
			Class.forName(callMainMethodClass).getMethod("main",String[].class).invoke(null,(Object)passThroughArguments);
		} catch (Exception e) {
			if(e instanceof ClassNotFoundException) { System.out.format("PBOSupport: provided Class %s could not be found.%n",callMainMethodClass); }
			if(e instanceof NoSuchMethodException) { System.out.format("PBOSupport: Class %s has no main method to call.%n", callMainMethodClass); }
			if(e instanceof SecurityException) { System.out.format("PBOSupport: PBOSupport doe snot have public access to Class %s or to its main method.%n", callMainMethodClass); }
			if(e instanceof IllegalArgumentException) { System.out.format("PBOSupport: There was an error in parsing out the main program's arguments.  Please immediately contact the maintainers of PbO.%n"); }
			if(e instanceof IllegalAccessException) { System.out.format("PBOSupport: Could not access the main method of Class %s.%n",callMainMethodClass); }
			if(e instanceof java.lang.reflect.InvocationTargetException) { throw e.getCause(); }
			
			e.printStackTrace();
			System.exit(-1);
		}
	
	}
}
END
	
	
} #end generateSupportFile

sub transformSource {

# transform source from $sourcename (assumes parameter declarations have been collected)
	for my $sourcename (@inputFiles) {
	    
	    my ($filename, $path) = fileparse($sourcename, $ext);
	    my $destName = "$path$filename.java";
		#my ($destName) = $sourcename =~ /(.+)$ext$/;
		
		print "Generating $destName.java from $sourcename ...\n";
		print $LOG "\nGenerating $destName.java from $sourcename ...\n";
	
		open( SRC,  "<$sourcename" );
		
		open( DEST, ">$destName" );
		
		# since we are moving PbOSupport.java to the
		# ca.ubc.beta.pbo package, we need a way
		# for source files to import this file
		# but this import MUST happen after an optional
		# package declaration
		#
		# these variables keep track of when that has happened
		my $seenPackageDeclaration = 0;
		my $importDeclared = 0;
	
		#print DEST "// generated from $sourcename by $weaverId\n\n";
	
		my $skipChoice     = 0;
		my $exitSkipChoice = 0;
		my $insertMain		= 0;
	
		my $k = 0;
		while (<SRC>) {
			$k++;
	
			# deal with parameter declarations (##PARAM):
			while (/$param_pattern/)
			{
				#$paramDeclLine{$4} = $k;
				printf $LOG ("\n  line $k: removed declaration of parameter ##$4\n");
				printf $LOG ("  changed line $k from: $_");
	
				# suppress line
				$_ =~
	s/$param_pattern//;
				printf $LOG ("    to: $_");
			}
	
	
	
	  # deal with choice point declarations (##BEGIN CHOICE, ##CHOICE, ##END CHOICE)
			while (/##BEGIN\s+CHOICE\s+(\w+)\s*=\s*(\w+)/) {
				#$choiceLines{$1}     = $choiceLines{$1} . "$k";
				#$choiceInstNames{$1} = $choiceInstNames{$1} . "\n    $2 (line $k)";
				
				my $oldln            = $_;
				
				# fixed for 0.6.05
				# need to map named choice instances to their mapParamVal equivalents
				if ( $mode eq "inst" && $instList{$1} eq mapParamVal($2) ) {
					$_ =~ s/##BEGIN\s+CHOICE\s+(\w+)\s*=(\w+)/\/\/ begin choice $1 [instantiated]/;
				}
				elsif ($mode eq "inst" && exists $instList{$1} && $instList{$1} ne mapParamVal($2) ) {
					$_ =~ s/##BEGIN\s+CHOICE\s+(\w+)\s*=(\w+)/\/\* begin choice $1 [instantiated]/;
					$skipChoice = 1;
				}
				else {
					$_ =~ s/##BEGIN\s+CHOICE\s+(\w+)\s*=(\w+)/if (PBOSupport.getInstance().$1==PBOSupport.getInstance().$1\_$2) \{ \/\/ begin choice $1/;
				}
				printf $LOG ("\n  line $k: resolved ##BEGIN CHOICE $1=$2\n");
				printf $LOG ("  changed line $k from: $oldln");
				printf $LOG ("    to: $_");
			}
			
			
			while (/##BEGIN\s+CHOICE\s+(\w+)/) {
				#$choiceLines{$1} = $choiceLines{$1} . "$k";
				
				my $oldln = $_;
				if ( $mode eq "inst" && $instList{$1} eq 1 ) {    # on
					$_ =~ s/##BEGIN\s+CHOICE\s+(\w+)/\/\/ begin choice $1 [instantiated]/;
				}
				elsif ( $mode eq "inst" && $instList{$1} eq 0 ) {    # off
					$_ =~ s/##BEGIN\s+CHOICE\s+(\w+)/\/\* begin choice $1 [instantiated]/;
					$skipChoice = 1;
				}
				else {
					$_ =~ s/##BEGIN\s+CHOICE\s+(\w+)/if (PBOSupport.getInstance().$1==1) \{ \/\/ begin choice $1/;
				}
				printf $LOG ("\n  line $k: resolved ##BEGIN CHOICE $1\n");
				printf $LOG ("  changed line $k from: $oldln");
				printf $LOG ("    to: $_");
			}
			
			
			while (/##CHOICE\s+(\w+)\s*=(\w+)/) {
				#$choiceInstNames{$1} = $choiceInstNames{$1} . "\n    $2 (line $k)";
				
				my $oldln = $_;
				if ( $mode eq "inst" && $instList{$1} eq $2 ) {
	
					# note: earlier instances must be stripped
					$_ =~
					  s/##CHOICE\s+(\w+)\s*=(\w+)/\/\/ choice $1=$2 [instantiated]/;
					$skipChoice = 0;
				}
				elsif ($mode eq "inst"
					&& exists $instList{$1}
					&& $instList{$1} ne $2 )
				{
	
	# note: earlier instances may be stripped or not; relaying here on actual stripping of the modified line
					$_ =~
					  s/##CHOICE\s+(\w+)\s*=(\w+)/\/\/ choice $1=$2 [instantiated]/;
					$skipChoice = 1;
				}
				else {
					$_ =~
	s/##CHOICE\s+(\w+)\s*=(\w+)/\} else if (PBOSupport.getInstance().$1==PBOSupport.getInstance().$1\_$2) \{ \/\/ choice $1=$2/;
				}
				printf $LOG ("\n  line $k: resolved ##CHOICE $1=$2\n");
				printf $LOG ("  changed line $k from: $oldln");
				printf $LOG ("    to: $_");
			}
			
			
			while (/##END\s+CHOICE\s+(\w+)/) {
				#$choiceLines{$1} = $choiceLines{$1} . "-$k ";
				
				my $oldln = $_;
				if ( $mode eq "inst" && exists $instList{$1} && !$skipChoice ) {
					$_ =~
					  s/##END\s+CHOICE\s+(\w+)/\/\/ end choice $1 [instantiated]/;
				}
				elsif ( $mode eq "inst" && exists $instList{$1} && $skipChoice ) {
					$_ =~
					  s/##END\s+CHOICE\s+(\w+)/  end choice $1 [instantiated] \*\//;
					$exitSkipChoice = 1;
				}
				else {
					$_ =~ s/##END\s+CHOICE\s+(\w+)/\} \/\/ end choice $1/;
				}
				printf $LOG ("\n  line $k: resolved ##END CHOICE $1\n");
				printf $LOG ("  changed line $k from: $oldln");
				printf $LOG ("    to: $_");
			}
	
			# added in 0.5.0
			while (/##LOG\s*\(([\w]+)\s+([\w]+)\s*\=\s*(.+)\)/ ) {
				#$logLines{$2} = $logLines{$2} . "$k ";
				
				printf $LOG ("\n line $k: resolved log statement (log $2)\n");
				printf $LOG ("  changed line $k from: $_");
				
				my $time = "(float)((System.currentTimeMillis()-PBOSupport.getInstance().startTime)/1000F)";
				
				my $val = &mapParamVal($3);
				$_ =~ s/##LOG\s*\(\s*([\w]+)\s+([\w]+)\s*\=\s*(.+)\)/System.out.format\(\"BEGIN LOG $2 [t=%f]\\n\%s\\nEND LOG $2%n\", $time, $val\)\;/;
			}
	
			# deal with parameter references (##<name>):
			while (/##((?!(?:PARAM|CHOICE|BEGIN|END))[\w]+)/) {
				#$paramAccLines{$1} = $paramAccLines{$1} . "$k ";
				
				my $oldln = $_;
				if ( $mode == "inst" && exists $instList{$1} ) {
					$_ =~ s/##([\w]+)/$instList{$1}/;
				}
				else {
					$_ =~ s/##([\w]+)/PBOSupport.getInstance().$1/;
				}
	
				printf $LOG ("\n  line $k: resolved parameter reference ##$1\n");
				printf $LOG ("  changed line $k from: $oldln");
				printf $LOG ("    to: $_");
			}
	
			if ( !$skipChoice ) {
			    
			    if (/^package/) { print DEST $_; print DEST "import ca.ubc.beta.pbo.PBOSupport;\n"; $importDeclared = 1;}
			    else { 
			        if ($importDeclared == 0) {
			            print DEST "import ca.ubc.beta.pbo.PBOSupport;\n$_";
			            $importDeclared = 1;
			        } else {
			            print DEST $_;
			        }
			    }
				
			}
			else {
				print DEST "\n";
			}
	
			if ($exitSkipChoice) {
				$skipChoice     = 0;
				$exitSkipChoice = 0;
			}
	
		}
	
		close(DEST);
	}
}

#sub generateInfoFile {

# generate .info file from parameter and choice information collected from source during transformation
#
#	open( INFO, ">pbo.info" );	
#	my $sourcename = join ", ", @inputFiles;
#	print INFO <<"END";
#// parameters and choices in $sourcename
#// generated by $weaverId
#
#// parameters:
#END
#
#	foreach my $key ( sort keys %paramAccLines ) {
#
#		# check whether parameter has been declared:
#		if ( exists $paramDeclLine{$key} ) {
#			printf INFO (
#"\n##$key:\n  declared in line: $paramDeclLine{$key}\n  used in line(s): $paramAccLines{$key}\n"
#			);
#		}
#		else {
#
#			# generate error message:
#			$paramAccLines{$key} =~ s/\s+$//;    # remove trailing white space
#			printf(
#"\nERROR: parameter $key, used in line(s) $paramAccLines{$key}, has not been declared.\n"
#			);
#			exit(-1);
#		}
#	}
#
#	printf INFO ("\n\n// choices:");
#
#	foreach my $key ( sort keys %choiceLines ) {
#		printf INFO ("\n\n##$key:\n  spans lines: $choiceLines{$key}");
#		if ( $choiceInstNames{$key} ne "" ) {
#			print INFO ("\n  includes instances: $choiceInstNames{$key}");
#		}
#	}
#
#	print INFO ("\n\n// EOF\n");
#
#	close(INFO);
#
#}

# ------------------ main program ------------------------------------------

# step 1: collect all param and choice declarations
#   currently assumes only one ##PARAM statement per line (TODO: lift restriction)
&collectCommandLineArguments(@ARGV);

open($LOG, ">$output_dir/pbo.log" );

printf("Weaving PbO-java files ".join(",",@inputFiles)." ...\n");
printf $LOG ("Weaving PbO-java files ". join(",",@inputFiles) ." with $weaverId ...\n\n");

&collectPBOConstructs;

#&collectParamDecl;
#&collectChoiceDecl;
#&getInfo;
# step 2: generate PBO_Support.java

&generateSupportFile;

# step 3: generate default configuration file

&generateConfigFile;

# step 4: transform source

&transformSource;

# step 5: generate .info file

&generateInfoFile;
&generatePropertiesFile;
&generateParamsFile;
&generateTranslationTable;

printf("\n... done.\n");

for my $sourcename (@inputFiles) {
	printf $LOG ("\n... finished weaving $sourcename.\n");
}

close($LOG);

# EOF
